@@ -606,9 +606,7 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	return PyString_FromStringAndSize((char *)&val, 1);
     }
     case G_TYPE_BOOLEAN: {
-	PyObject *val = g_value_get_boolean(value) ? Py_True : Py_False;
-	Py_INCREF(val);
-	return val;
+	return PyBool_FromLong(g_value_get_boolean(value));
     }
     case G_TYPE_INT:
 	return PyInt_FromLong(g_value_get_int(value));
@@ -723,15 +721,17 @@ static void
 pyg_closure_invalidate(gpointer data, GClosure *closure)
 {
     PyGClosure *pc = (PyGClosure *)closure;
+    PyGILState_STATE state;
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
     Py_XDECREF(pc->callback);
     Py_XDECREF(pc->extra_args);
     Py_XDECREF(pc->swap_data);
+    pyg_gil_state_release(state);
+
     pc->callback = NULL;
     pc->extra_args = NULL;
     pc->swap_data = NULL;
-    pyg_unblock_threads();
 }
 
 static void
@@ -747,8 +747,7 @@ pyg_closure_marshal(GClosure *closure,
     PyObject *params, *ret;
     guint i;
 
-    pyg_block_threads();
-    state = PyGILState_Ensure();
+    state = pyg_gil_state_ensure();
 
     /* construct Python tuple for the parameter values */
     params = PyTuple_New(n_param_values);
@@ -786,8 +785,7 @@ pyg_closure_marshal(GClosure *closure,
  out:
     Py_DECREF(params);
     
-    PyGILState_Release(state);
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 }
 
 /**
@@ -849,6 +847,7 @@ pyg_signal_class_closure_marshal(GClosure *closure,
 				 gpointer invocation_hint,
 				 gpointer marshal_data)
 {
+    PyGILState_STATE state;
     GObject *object;
     PyObject *object_wrapper;
     GSignalInvocationHint *hint = (GSignalInvocationHint *)invocation_hint;
@@ -857,9 +856,9 @@ pyg_signal_class_closure_marshal(GClosure *closure,
     PyObject *params, *ret;
     guint i, len;
 
+    state = pyg_gil_state_ensure();
+    
     g_return_if_fail(invocation_hint != NULL);
-
-    pyg_block_threads();
     /* get the object passed as the first argument to the closure */
     object = g_value_get_object(&param_values[0]);
     g_return_if_fail(object != NULL && G_IS_OBJECT(object));
@@ -883,7 +882,7 @@ pyg_signal_class_closure_marshal(GClosure *closure,
     if (!method) {
 	PyErr_Clear();
 	Py_DECREF(object_wrapper);
-	pyg_unblock_threads();
+	pyg_gil_state_release(state);
 	return;
     }
     Py_DECREF(object_wrapper);
@@ -897,7 +896,7 @@ pyg_signal_class_closure_marshal(GClosure *closure,
 	/* error condition */
 	if (!item) {
 	    Py_DECREF(params);
-	    pyg_unblock_threads();
+	    pyg_gil_state_release(state);
 	    return;
 	}
 	PyTuple_SetItem(params, i - 1, item);
@@ -924,7 +923,7 @@ pyg_signal_class_closure_marshal(GClosure *closure,
 	PyErr_Print();
 	Py_DECREF(method);
 	Py_DECREF(params);
-	pyg_unblock_threads();
+	pyg_gil_state_release(state);
 	return;
     }
     Py_DECREF(method);
@@ -932,7 +931,7 @@ pyg_signal_class_closure_marshal(GClosure *closure,
     if (return_value)
 	pyg_value_from_pyobject(return_value, ret);
     Py_DECREF(ret);
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 }
 
 /**
