@@ -1,11 +1,30 @@
-/* -*- Mode: C; c-basic-offset: 4 -*- */
+/* -*- Mode: C; c-basic-offset: 4 -*-
+ * pygtk- Python bindings for the GTK toolkit.
+ * Copyright (C) 1998-2003  James Henstridge
+ *
+ *   pygtktreemodel.c: stub class to help implement tree models.
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * USA
+ */
 #ifdef HAVE_CONFIG_H
 #  include "config.h"
 #endif
 
 #include "pygtktreemodel.h"
 #include <Python.h>
-#include "pygobject.h"
 #include "pygtk-private.h"
 
 /* define this to print out debug messages */
@@ -15,6 +34,9 @@
 #  define _(s) (s)
 #endif
 
+#define VALID_ITER(iter, tree_model) \
+ (iter != NULL && iter->stamp == PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp)
+
 enum {
     PROP_LEAK_REFERENCES = 1
 };
@@ -113,6 +135,10 @@ static gboolean pygtk_generic_tree_model_iter_nth_child(GtkTreeModel *tree_model
 static gboolean pygtk_generic_tree_model_iter_parent(GtkTreeModel *tree_model,
 					     GtkTreeIter *iter,
 					     GtkTreeIter *child);
+static void pygtk_generic_tree_model_unref_node(GtkTreeModel *tree_model,
+					     GtkTreeIter *iter);
+static void pygtk_generic_tree_model_ref_node(GtkTreeModel *tree_model,
+					     GtkTreeIter *iter);
 
 static void
 pygtk_generic_tree_model_iface_init(GtkTreeModelIface *iface)
@@ -129,12 +155,18 @@ pygtk_generic_tree_model_iface_init(GtkTreeModelIface *iface)
   iface->iter_n_children = pygtk_generic_tree_model_iter_n_children;
   iface->iter_nth_child = pygtk_generic_tree_model_iter_nth_child;
   iface->iter_parent = pygtk_generic_tree_model_iter_parent;
+  iface->ref_node = pygtk_generic_tree_model_ref_node;
+  iface->unref_node = pygtk_generic_tree_model_unref_node;
+
 }
 
 static void
 pygtk_generic_tree_model_init(PyGtkGenericTreeModel *self)
 {
     self->leak_references = TRUE;
+    do {
+        self->stamp = g_random_int();
+    } while (self->stamp == 0);
 }
 
 static void
@@ -188,97 +220,107 @@ pygtk_generic_tree_model_new(void)
 static guint
 pygtk_generic_tree_model_get_flags(GtkTreeModel *tree_model)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret;
-
+    guint ret = 0;
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), 0);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("get_flags()");
 #endif
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_flags", "");
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
-	guint ret = PyInt_AsLong(py_ret);
-
+	ret = PyInt_AsLong(py_ret);
 	Py_DECREF(py_ret);
-	return ret;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return 0;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gint
 pygtk_generic_tree_model_get_n_columns(GtkTreeModel *tree_model)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret;
-
+    gint ret = 0;
+    
     g_return_val_if_fail(tree_model != NULL, 0);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), 0);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("get_n_columns()");
 #endif
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_n_columns", "");
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
-	gint ret = PyInt_AsLong(py_ret);
-
+	ret = PyInt_AsLong(py_ret);
 	Py_DECREF(py_ret);
-	return ret;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return 0;
     }
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static GType
 pygtk_generic_tree_model_get_column_type(GtkTreeModel *tree_model, gint index)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret;
-
+    GType ret = G_TYPE_INVALID;
+    
     g_return_val_if_fail(tree_model != NULL, G_TYPE_INVALID);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), G_TYPE_INVALID);
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("get_column_type(%d)", index);
 #endif
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_column_type",
 				 "(i)", index);
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
-	GType ret = pyg_type_from_object(py_ret);
-
+	ret = pyg_type_from_object(py_ret);
 	Py_DECREF(py_ret);
-	return ret;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return G_TYPE_INVALID;
     }
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gboolean
 pygtk_generic_tree_model_get_iter(GtkTreeModel *tree_model,
-			  GtkTreeIter *iter, GtkTreePath *path)
+                                  GtkTreeIter *iter, GtkTreePath *path)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_path, *py_ret;
-
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
     g_return_val_if_fail(iter != NULL, FALSE);
     g_return_val_if_fail(path != NULL, FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
@@ -286,74 +328,86 @@ pygtk_generic_tree_model_get_iter(GtkTreeModel *tree_model,
     g_message("get_iter(%p)", path);
 #endif
     py_path = pygtk_tree_path_to_pyobject(path);
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_iter",
 				 "(O)", py_path);
-    pyg_unblock_threads();
+    Py_DECREF(self);
     Py_DECREF(py_path);
 
     if (py_ret) {
 	if (py_ret != Py_None) {
 	    iter->user_data = py_ret;
+            iter->stamp = PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp;
 	    if (!PYGTK_GENERIC_TREE_MODEL(tree_model)->leak_references) {
 		Py_DECREF((PyObject *)iter->user_data);
 	    }
-	    return TRUE;
+	    ret = TRUE;
 	} else {
 	    iter->user_data = NULL;
 	    Py_DECREF(py_ret);
-	    return FALSE;
 	}
     } else {
 	PyErr_Print();
-	PyErr_Clear();
 	iter->user_data = NULL;
-	return FALSE;
     }
+
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static GtkTreePath *
 pygtk_generic_tree_model_get_path(GtkTreeModel *tree_model, GtkTreeIter *iter)
 {
-    PyObject *self, *py_ret;
-
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter;
+    GtkTreePath *path = NULL;
+    
     g_return_val_if_fail(tree_model != NULL, NULL);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), NULL);
-    g_return_val_if_fail(iter != NULL, NULL);
+    g_return_val_if_fail(VALID_ITER(iter, tree_model), NULL);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("get_path(%p)", iter);
 #endif
-    pyg_block_threads();
-    py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_path",
-				 "(O)", (PyObject *)iter->user_data);
-    pyg_unblock_threads();
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
+
+    py_ret = PyObject_CallMethod(self, METHOD_PREFIX "get_path", "(O)",
+                                 py_iter);
+    Py_DECREF(self);
     if (py_ret) {
-	GtkTreePath *path = pygtk_tree_path_from_pyobject(py_ret);
+	path = pygtk_tree_path_from_pyobject(py_ret);
 
 	if (!path)
 	    g_warning("could not convert return value of get_path() to "
 		      "a GtkTreePath");
 	Py_DECREF(py_ret);
-	return path;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return NULL;
     }
+    
+    pyg_gil_state_release(state);
+    return path;
 }
 
 static void
 pygtk_generic_tree_model_get_value(GtkTreeModel*tree_model, GtkTreeIter *iter,
-			   gint column, GValue *value)
+                                   gint column, GValue *value)
 {
-    PyObject *self, *py_value;
+    PyGILState_STATE state;
+    PyObject *self, *py_value, *py_iter;
 
     g_return_if_fail(tree_model != NULL);
     g_return_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model));
-    g_return_if_fail(iter != NULL);
+    g_return_if_fail(VALID_ITER(iter, tree_model));
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
@@ -362,40 +416,52 @@ pygtk_generic_tree_model_get_value(GtkTreeModel*tree_model, GtkTreeIter *iter,
     _PyObject_Dump (iter->user_data);
 #endif
     /* init value to column type */
-    g_value_init(value, pygtk_generic_tree_model_get_column_type(tree_model, column));
+    g_value_init(value, pygtk_generic_tree_model_get_column_type(tree_model,
+                                                                 column));
 
-    pyg_block_threads();
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
     py_value = PyObject_CallMethod(self, METHOD_PREFIX "get_value",
-				   "(Oi)", (PyObject *)iter->user_data,column);
-    pyg_unblock_threads();
+				   "(Oi)", py_iter,column);
+    Py_DECREF(self);
 
     if (py_value) {
-	pyg_value_from_pyobject(value, py_value);
-	Py_DECREF(py_value);
+        if (py_value != Py_None)
+            pyg_value_from_pyobject(value, py_value);
+        Py_DECREF(py_value);
     } else {
 	PyErr_Print();
-	PyErr_Clear();
     }
+    pyg_gil_state_release(state);
 }
 
 static gboolean
 pygtk_generic_tree_model_iter_next(GtkTreeModel *tree_model, GtkTreeIter *iter)
 {
-    PyObject *self, *py_ret;
-
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter;
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
-    g_return_val_if_fail(iter != NULL, FALSE);
+    g_return_val_if_fail(VALID_ITER(iter, tree_model), FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("iter_next(%p)", iter);
 #endif
-    pyg_block_threads();
-    py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_next",
-				 "(O)", (PyObject *)iter->user_data);
-    pyg_unblock_threads();
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
+
+    py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_next", "(O)",
+                                 py_iter);
+    Py_DECREF(self);
     if (py_ret) {
 	if (py_ret != Py_None) {
 	    /* XXXX handle reference counting here */
@@ -403,29 +469,36 @@ pygtk_generic_tree_model_iter_next(GtkTreeModel *tree_model, GtkTreeIter *iter)
 	    if (!PYGTK_GENERIC_TREE_MODEL(tree_model)->leak_references) {
 		Py_DECREF((PyObject *)iter->user_data);
 	    }
-	    return TRUE;
+	    ret = TRUE;
 	} else {
 	    iter->user_data = NULL;
 	    Py_DECREF(py_ret);
-	    return FALSE;
 	}
     } else {
-	iter->user_data = NULL;
+        iter->user_data = NULL;
 	PyErr_Print();
-	PyErr_Clear();
-	return FALSE;
     }
+
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gboolean
-pygtk_generic_tree_model_iter_children(GtkTreeModel *tree_model, GtkTreeIter *iter,
-			       GtkTreeIter *parent)
+pygtk_generic_tree_model_iter_children(GtkTreeModel *tree_model,
+                                       GtkTreeIter *iter,
+                                       GtkTreeIter *parent)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret, *py_parent = Py_None;
-
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
     g_return_val_if_fail(iter != NULL, FALSE);
+    g_return_val_if_fail(parent == NULL || parent->stamp == PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp, FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
@@ -434,100 +507,124 @@ pygtk_generic_tree_model_iter_children(GtkTreeModel *tree_model, GtkTreeIter *it
 #endif
     if (parent && parent->user_data != NULL)
 	py_parent = (PyObject *)parent->user_data;
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_children",
 				 "(O)", py_parent);
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
 	if (py_ret != Py_None) {
 	    /* XXXX handle reference counting here */
 	    iter->user_data = py_ret;
+            iter->stamp = PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp;
 	    if (!PYGTK_GENERIC_TREE_MODEL(tree_model)->leak_references) {
 		Py_DECREF((PyObject *)iter->user_data);
 	    }
-	    return TRUE;
+	    ret = TRUE;
 	} else {
 	    iter->user_data = NULL;
 	    Py_DECREF(py_ret);
-	    return FALSE;
 	}
     } else {
 	iter->user_data = NULL;
 	PyErr_Print();
-	PyErr_Clear();
-	return FALSE;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gboolean
-pygtk_generic_tree_model_iter_has_child(GtkTreeModel *tree_model, GtkTreeIter *iter)
+pygtk_generic_tree_model_iter_has_child(GtkTreeModel *tree_model,
+                                        GtkTreeIter *iter)
 {
-    PyObject *self, *py_ret;
-
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter;
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
-    g_return_val_if_fail(iter != NULL, FALSE);
+    g_return_val_if_fail(VALID_ITER(iter, tree_model), FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("iter_has_child(%p)", iter);
 #endif
-    pyg_block_threads();
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
+
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_has_child",
-				 "(O)", (PyObject *)iter->user_data);
-    pyg_unblock_threads();
+				 "(O)", py_iter);
+    Py_DECREF(self);
     if (py_ret) {
-	gboolean ret = PyObject_IsTrue(py_ret);
+	ret = PyObject_IsTrue(py_ret);
 
 	Py_DECREF(py_ret);
-	return ret;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return FALSE;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gint
-pygtk_generic_tree_model_iter_n_children(GtkTreeModel *tree_model, GtkTreeIter *iter)
+pygtk_generic_tree_model_iter_n_children(GtkTreeModel *tree_model,
+                                         GtkTreeIter *iter)
 {
-    PyObject *self, *py_ret;
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter;
+    guint ret = FALSE;
+    
+    g_return_val_if_fail(tree_model != NULL, 0);
+    g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), 0);
+    g_return_val_if_fail(iter == NULL || iter->stamp == PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp, 0);
+
+    state = pyg_gil_state_ensure();
 
-    g_return_val_if_fail(tree_model != NULL, FALSE);
-    g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
-    g_return_val_if_fail(iter != NULL, FALSE);
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
 #ifdef DEBUG_TREE_MODEL
     g_message("iter_n_children(%p)", iter);
 #endif
-    pyg_block_threads();
+
+    py_iter = iter != NULL ? (PyObject *)iter->user_data : Py_None;
+
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_n_children",
-				 "(O)", (PyObject *)iter->user_data);
-    pyg_unblock_threads();
+				 "(O)", py_iter);
+    Py_DECREF(self);
     if (py_ret) {
-	gint ret = PyInt_AsLong(py_ret);
+	ret = PyInt_AsLong(py_ret);
 
 	Py_DECREF(py_ret);
-	return ret;
     } else {
 	PyErr_Print();
-	PyErr_Clear();
-	return 0;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gboolean
-pygtk_generic_tree_model_iter_nth_child(GtkTreeModel *tree_model, GtkTreeIter  *iter,
-				GtkTreeIter  *parent, gint n)
+pygtk_generic_tree_model_iter_nth_child(GtkTreeModel *tree_model,
+                                        GtkTreeIter  *iter,
+                                        GtkTreeIter  *parent, gint n)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret, *py_parent = Py_None;
-
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
     g_return_val_if_fail(iter != NULL, FALSE);
+    g_return_val_if_fail(parent == NULL || parent->stamp == PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp, FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
@@ -536,40 +633,47 @@ pygtk_generic_tree_model_iter_nth_child(GtkTreeModel *tree_model, GtkTreeIter  *
 #endif
     if (parent && parent->user_data != NULL)
 	py_parent = (PyObject *)parent->user_data;
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_nth_child",
 				 "(Oi)", py_parent, n);
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
 	if (py_ret != Py_None) {
 	    /* XXXX handle reference counting here */
 	    iter->user_data = py_ret;
+            iter->stamp = PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp;
 	    if (!PYGTK_GENERIC_TREE_MODEL(tree_model)->leak_references) {
 		Py_DECREF((PyObject *)iter->user_data);
 	    }
-	    return TRUE;
+	    ret = TRUE;
 	} else {
 	    iter->user_data = NULL;
 	    Py_DECREF(py_ret);
-	    return FALSE;
 	}
     } else {
 	iter->user_data = NULL;
 	PyErr_Print();
-	PyErr_Clear();
-	return FALSE;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
 static gboolean
-pygtk_generic_tree_model_iter_parent(GtkTreeModel *tree_model, GtkTreeIter *iter,
-			     GtkTreeIter *child)
+pygtk_generic_tree_model_iter_parent(GtkTreeModel *tree_model,
+                                     GtkTreeIter *iter,
+                                     GtkTreeIter *child)
 {
+    PyGILState_STATE state;
     PyObject *self, *py_ret, *py_child = Py_None;
-
+    gboolean ret = FALSE;
+    
     g_return_val_if_fail(tree_model != NULL, FALSE);
     g_return_val_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model), FALSE);
     g_return_val_if_fail(iter != NULL, FALSE);
+    g_return_val_if_fail(VALID_ITER(child, tree_model), FALSE);
+
+    state = pyg_gil_state_ensure();
+
     /* this call finds the wrapper for this GObject */
     self = pygobject_new((GObject *)tree_model);
 
@@ -578,28 +682,122 @@ pygtk_generic_tree_model_iter_parent(GtkTreeModel *tree_model, GtkTreeIter *iter
 #endif
     if (child && child->user_data != NULL)
 	py_child = (PyObject *)child->user_data;
-    pyg_block_threads();
     py_ret = PyObject_CallMethod(self, METHOD_PREFIX "iter_parent",
 				 "(O)", py_child);
-    pyg_unblock_threads();
+    Py_DECREF(self);
     if (py_ret) {
 	if (py_ret != Py_None) {
 	    /* XXXX handle reference counting here */
 	    iter->user_data = py_ret;
+            iter->stamp = PYGTK_GENERIC_TREE_MODEL(tree_model)->stamp;
 	    if (!PYGTK_GENERIC_TREE_MODEL(tree_model)->leak_references) {
 		Py_DECREF((PyObject *)iter->user_data);
 	    }
-	    return TRUE;
+	    ret = TRUE;
 	} else {
 	    iter->user_data = NULL;
 	    Py_DECREF(py_ret);
-	    return FALSE;
 	}
     } else {
 	iter->user_data = NULL;
 	PyErr_Print();
-	PyErr_Clear();
-	return FALSE;
     }
+    
+    pyg_gil_state_release(state);
+    return ret;
 }
 
+static void
+pygtk_generic_tree_model_unref_node(GtkTreeModel *tree_model, GtkTreeIter *iter)
+{
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter, *method;
+
+    g_return_if_fail(tree_model != NULL);
+    g_return_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model));
+    g_return_if_fail(VALID_ITER(iter, tree_model));
+
+    state = pyg_gil_state_ensure();
+
+    /* this call finds the wrapper for this GObject */
+    self = pygobject_new((GObject *)tree_model);
+
+#ifdef DEBUG_TREE_MODEL
+    g_message("unref_node(%p)", iter);
+#endif
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
+
+    method = PyObject_GetAttrString(self, METHOD_PREFIX "unref_node");
+    if (method == NULL)
+	PyErr_Clear();
+    else {
+	py_ret = PyObject_CallFunction(method, "(O)", py_iter);
+	if (py_ret) {
+	    Py_DECREF(py_ret);
+	} else {
+	    PyErr_Print();
+	}
+    }
+    Py_DECREF(self);
+
+    pyg_gil_state_release(state);
+}
+
+static void
+pygtk_generic_tree_model_ref_node(GtkTreeModel *tree_model, GtkTreeIter *iter)
+{
+    PyGILState_STATE state;
+    PyObject *self, *py_ret, *py_iter, *method;
+
+    g_return_if_fail(tree_model != NULL);
+    g_return_if_fail(PYGTK_IS_GENERIC_TREE_MODEL(tree_model));
+    g_return_if_fail(VALID_ITER(iter, tree_model));
+
+    state = pyg_gil_state_ensure();
+
+    /* this call finds the wrapper for this GObject */
+    self = pygobject_new((GObject *)tree_model);
+
+#ifdef DEBUG_TREE_MODEL
+    g_message("ref_node(%p)", iter);
+#endif
+    py_iter = (PyObject *)iter->user_data;
+    if (py_iter == NULL)
+	py_iter = Py_None;
+
+    method = PyObject_GetAttrString(self, METHOD_PREFIX "ref_node");
+    if (method == NULL)
+	PyErr_Clear();
+    else {
+	py_ret = PyObject_CallFunction(method, "(O)", py_iter);
+	if (py_ret) {
+	    Py_DECREF(py_ret);
+	} else {
+	    PyErr_Print();
+	}
+    }
+    Py_DECREF(self);
+
+    pyg_gil_state_release(state);
+}
+
+void
+pygtk_generic_tree_model_invalidate_iters(PyGtkGenericTreeModel *tree_model)
+{
+    g_return_if_fail(tree_model != NULL);
+
+    do {
+        tree_model->stamp++;
+    } while (tree_model->stamp == 0);
+}
+
+gboolean
+pygtk_generic_tree_model_iter_is_valid(PyGtkGenericTreeModel *tree_model,
+                                       GtkTreeIter *iter)
+{
+    g_return_val_if_fail(tree_model != NULL, FALSE);
+
+    return VALID_ITER(iter, tree_model);
+}
